var querystring= require('querystring'),
    crypto= require('crypto'),
    https= require('https'),
    http= require('http'),
    querystring = require('querystring'),
    URL= require('url'),
    OAuthUtils= require('./_utils');

exports.OAuth2= function(clientId, clientSecret, baseSite, authorizePath, accessTokenPath) {
  this._clientId= clientId;
  this._clientSecret= clientSecret; 
  this._baseSite= baseSite;
  this._authorizeUrl= authorizePath || "/oauth/authorize";
  this._accessTokenUrl= accessTokenPath || "/oauth/access_token";
  this._accessTokenName= "access_token";
}

// This 'hack' method is required for sites that don't use
// 'access_token' as the name of the access token (for requests).
// ( http://tools.ietf.org/html/draft-ietf-oauth-v2-16#section-7 )
// it isn't clear what the correct value should be atm, so allowing
// for specific (temporary?) override for now.
exports.OAuth2.prototype.setAccessTokenName= function ( name ) {
  this._accessTokenName= name;
}

exports.OAuth2.prototype._getAccessTokenUrl= function() {
  return this._baseSite + this._accessTokenUrl; /* + "?" + querystring.stringify(params); */ //JS HACK TO MAKE WORK WITH M WEST. TODO: look up if this is standard at all.
}

//TODO JS take in the content-type header as a parameter (similar to oauth 1 version)
//default to application/json as M requires that
exports.OAuth2.prototype._request= function(method, url, headers, post_body, access_token, callback) {

  var creds = crypto.createCredentials({ });  
  var parsedUrl= URL.parse( url, true );   
  if( parsedUrl.protocol == "https:" && !parsedUrl.port ) parsedUrl.port= 443;
  
  var realHeaders= {};
  if( headers ) {
    for(var key in headers) {
      realHeaders[key] = headers[key];
    }
  }
  realHeaders['Host']= parsedUrl.host;
  realHeaders['Authorization'] = "bearer " + access_token; //JS HACK TO MAKE WORK WITH M WEST. TODO: look up if this is standard at all.
  realHeaders['Accept'] = "*/*"; //JS HACK TO MAKE WORK WITH M WEST. TODO: look up if this is standard at all.
  realHeaders['Content-Type'] = "application/json; charset=utf8"; //JS HACK TO MAKE WORK WITH M WEST. TODO: look up if this is standard at all.
  
  realHeaders['Content-Length']= post_body ? Buffer.byteLength(post_body) : 0;
  if( access_token ) {
    if( ! parsedUrl.query ) parsedUrl.query= {};
    //parsedUrl.query[this._accessTokenName]= access_token; //JS HACK TO MAKE WORK WITH M WEST. TODO: look up if this is standard at all.
  }

  var result= "";
  var queryStr= querystring.stringify(parsedUrl.query);
  if( queryStr ) queryStr=  "?" + queryStr;
  var options = {
    host:parsedUrl.hostname,
    port: parsedUrl.port,
    path: parsedUrl.pathname + queryStr,
    method: method,
    headers: realHeaders
  };

  // Some hosts *cough* google appear to close the connection early / send no content-length header
  // allow this behaviour.
  var allowEarlyClose= OAuthUtils.isAnEarlyCloseHost(options.host);
  var callbackCalled= false;
  function passBackControl( response, result ) {
    if(!callbackCalled) {
      callbackCalled=true;
      if( response.statusCode > 399 ) { //JS HACK to fix to work with M API which may return 201 or redirects (300+)
        callback({ statusCode: response.statusCode, data: result });
      } else {
        callback(null, result, response);
      }
    }
  }
  

  var request = https.request(options, function (response) {
    response.on("data", function (chunk) {
      result+= chunk
    });
    response.on("close", function (err) {
      if( allowEarlyClose ) {
        passBackControl( response, result );
      }
    });
    response.addListener("end", function () {
      passBackControl( response, result );
    });
  });
  
  request.socket.setTimeout(1000); //JS HACK to get to work with M api. KNOWN socket timeouts on debit endpoints
  
  //The following is a JS HACK to get to work with M api.
  //for certain bank api debit endpoints the socket's will just
  //timeout if there is no data available. So we catch the timeout and
  //send it through to the call back so that the request can be restarted.
  request.socket.addListener("timeout", function() {
	  request.socket.destroy();
	  console.error("Socket Timeout. Connection Cosed");
	  callbackCalled= true;
	  callback("socket_timeout");
  });
  
  //console.log(require('sys').inspect(request));
  
  request.on('error', function(e) {
    callbackCalled= true;
    callback(e);
  });

  if(  method == 'POST' && post_body ) {
	  //('sending post body:'+post_body);
     request.write(post_body);
  }
  request.end();
} 

/*
 * JS HACK this is supposed to be a drop in replacement for the oauth1 version
 * we simply ignore the params we don't need
 */
exports.OAuth2.prototype._performSecureRequest= function( oauth_token, oauth_secret, method, url, extra_params, post_body, post_content_type, callback ) {
	if( (method == "POST" || method == "PUT")  && ( post_body == null && extra_params != null) ) {
	    //post_body= querystring.stringify(extra_params);
		post_body = JSON.stringify(extra_params);
	}
	
	this._request(method, url, null, post_body, oauth_token, callback);
}

/*
 * JS HACK this is supposed to be a drop in replacement for the oauth1 version
 * we simply ignore the params we don't need
 */
exports.OAuth2.prototype._putOrPost= function(method, url, oauth_token, oauth_secret, post_body, post_content_type, callback) {
  var extra_params= null;
  if( typeof post_content_type == "function" ) {
    callback= post_content_type;
    post_content_type= null;
  }
  if( typeof post_body != "string" ) {
    post_content_type= "application/x-www-form-urlencoded"
    extra_params= post_body;
    post_body= null;
  }
  this._performSecureRequest( oauth_token, null, method, url, extra_params, post_body, null, callback );
}

exports.OAuth2.prototype.getAuthorizeUrl= function( params ) {
  var params= params || {};
  params['client_id'] = this._clientId;
  params['type'] = 'web_server';
  return this._baseSite + this._authorizeUrl + "?" + querystring.stringify(params);
}

exports.OAuth2.prototype.getOAuthAccessToken= function(code, params, callback) {
  var params= params || {};
  params['client_id'] = this._clientId;
  params['client_secret'] = this._clientSecret;
  params['type']= 'web_server';
  params['code']= code;

  var post_data= querystring.stringify( params );
  var post_headers= {
       'Content-Type': 'application/x-www-form-urlencoded'
   };


  this._request("POST", this._getAccessTokenUrl(), post_headers, post_data, null, function(error, data, response) {
    if( error )  callback(error);
    else {
      var results;
      try {
        // As of http://tools.ietf.org/html/draft-ietf-oauth-v2-07
        // responses should be in JSON
        results= JSON.parse( data );
      }
      catch(e) {
        // .... However both Facebook + Github currently use rev05 of the spec
        // and neither seem to specify a content-type correctly in their response headers :(
        // clients of these services will suffer a *minor* performance cost of the exception
        // being thrown
        results= querystring.parse( data );
      }
      var access_token= results["access_token"];
      var refresh_token= results["refresh_token"];
      delete results["refresh_token"];
      callback(null, access_token, refresh_token);
    }
  });
} 

// Deprecated
exports.OAuth2.prototype.getProtectedResource= function(url, access_token, callback) {
  this._request("GET", url, {}, "", access_token, callback );
}

exports.OAuth2.prototype.get= function(url, access_token, callback) {
  this._request("GET", url, {}, "", access_token, callback );
}

/*
 * JS HACK this is supposed to be similar to the oauth1 version
 * method signature differs because we don't need the token_secret
 * TODO allow for the post content-type to be passed in. for now 
 * we are ignoring it and always sending application/json
 * as M requires that. but not all servers handle only
 * application/json
 */
exports.OAuth2.prototype.put= function(url, access_token, post_body, callback) {
  this._putOrPost("PUT", url, access_token, null, post_body, null, callback);
}

/*
 * JS HACK this is supposed to be a drop in replacement for the oauth1 version
 * we simply ignore the params we don't need
 */
exports.OAuth2.prototype.post= function(url, access_token, post_body, callback) {
  this._putOrPost("POST", url, access_token, null, post_body, null, callback);
}
